from .output import Output
import os
import gevent
import subprocess
import datetime
from ..event import Event
from ..util import Periodic

class File(Output):
    """Log to file.

    The file format is a line per event as json.

    When a log is rolled, a 'virtual' event will be generated with the tag
    'fileroll', and the field 'filename' which can be used by further outputs to
    process a log file when it rolls (eg. batch upload to S3). This event contains
    a 'trigger' field containing the original event that caused the log roll.

    :param string filename: the log filename (required). You can use event
      format values in this (eg. 'output-{program}.log')
    :param integer max_size: maximum size of file before rolling to .1, .2, etc.
    :param integer max_count: maximum number of rolled files (default: 10)
    :param string compress: set to 'gz' to compress the file after rolling.
    """
    def __init__(self, filename, max_size=None, max_count=10, compress=None):
        super(File, self).__init__()
        self.filename = filename
        self.max_size = max_size
        self.max_count = max_count
        if compress is True:
            compress = 'gz'
        elif compress is None:
            compress = False
        assert compress in (False, 'gz',)
        self.compress = compress
        self.last_filename = None
        self.last_event = None

        if 'timestamp' in self.filename:
            # if the log file name is timestamped, periodically check if it has
            # rolled over.
            self.periodic = Periodic(30, self._check_rotate)
        else:
            self.periodic = None

    def _check_rotate(self):
        now = datetime.datetime.utcnow()
        filename = self.filename.format(timestamp=now)
        if self.last_filename and filename != self.last_filename:
            self._rotate(self.last_filename, self.last_event, self.last_event)
            self.last_filename = filename 

    def start(self):
        super(File, self).start()
        if self.periodic is not None:
            self.periodic.start()

    def stop(self):
        if self.periodic is not None:
            self.periodic.kill()
            self.periodic.join()
        super(File, self).stop()

    def process(self, event):
        filename = event.format(self.filename)
        if self.max_size and os.path.exists(filename) and os.path.getsize(filename) > self.max_size:
            self._rotate(filename, self.last_event, event)

        if 'timestamp' in self.filename and self.last_filename != filename:
            if self.last_filename:
                # file has rolled by timestamp
                self._rotate(self.last_filename, self.last_event, event)
            self.last_filename = filename
        self.last_event = event

        dirname = os.path.dirname(filename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with file(filename, 'a') as fout:
            print >>fout, event.to_json()

    def _rotate(self, filename, last, trigger):
        if not os.path.exists(filename):
            return

        suffix = ''
        if self.compress == 'gz':
            suffix = '.'+self.compress

        roll_first = filename+'.1'+suffix
        renames = [(filename, roll_first)]
        self.logger.info('Rotating logfile %s to %s' % (filename, roll_first))
        for n, (oldname, newname) in enumerate(renames):
            if self.max_count and n == self.max_count-1:
                # cap renamed files if max_count specified
                break

            if os.path.exists(newname):
                base, num = newname.strip(suffix).rsplit('.', 1)
                renames.append((newname, '%s.%s%s' % (base, int(num)+1, suffix)))

        if suffix:
            # actually a rename, then compress
            renames[0] = (filename, filename+'.1')

        for oldname, newname in reversed(renames):
            self.logger.debug('Renaming logfile %s->%s' % (oldname, newname))
            os.rename(oldname, newname)

        if self.compress == 'gz':
            self._gz(filename+'.1')

        # emit 'virtual' event for rolled log file
        self.output.put(Event(tags=['fileroll'], filename=roll_first, last=last or trigger, trigger=trigger))

    def _gz(self, filename):
        self.logger.debug('Gzipping %s' % (filename,))
        ps = subprocess.Popen(['gzip', filename])
        while ps.poll() is None:
            gevent.sleep(0.01)



import types

from .filter import Filter

class Mutate(Filter):
    """Filter that allows you to add, rename, copy and drop fields

    :param map set: fields to set (optional). The values if strings may format other fields from the event.
    :param map rename: fields to rename (a: b renames b to a) (optional)
    :param list unset: fields to unset (optional)
    """
    def __init__(self, set={}, rename={}, copy={}, unset=[]):
        super(Mutate, self).__init__()
        self.sets = set
        assert type(self.sets) == dict
        self.renames = rename
        assert type(self.renames) == dict
        self.copies = copy
        assert type(self.copies) == dict
        self.unsets = unset
        assert type(self.unsets) == list
    
    def process(self, event):
        for k, v in self.sets.iteritems():
            if isinstance(v, types.StringTypes):
                v = event.format(v)
            event[k] = v
            self.logger.debug('Set %r to %r' % (k, v))

        for k, v in self.renames.iteritems():
            if v in event:
                event[k] = event.pop(v)
                self.logger.debug('Renamed %r to %r' % (v, k))

        for k, v in self.copies.iteritems():
            if v in event:
                event[k] = event[v]
                self.logger.debug('Copied %r to %r' % (v, k))

        for k in self.unsets:
            if k in event:
                del event[k]
                self.logger.debug('Unset %r' % (k,))



#!/usr/bin/env python

import logcabin
import sys
import os
import json
import gevent
import gevent.event
import gevent.monkey
gevent.monkey.patch_all()
import logging
import logging.config
import signal
import optparse
from util import ConfigException
from configuration import PyConfigLoader

class LogCabin(object):
    """Main application object"""

    def setup(self):
        """Application setup"""
        self._setup_args()
        self._setup_logging()
        self._setup_config()
        self._setup_signals()

    def _setup_logging(self):
        # configure logging
        logging.config.dictConfig(self.logconfig)
        self.logger = logging.getLogger('main')

    def _setup_config(self):
        # load configuration
        self.config = PyConfigLoader(self.opts.config)
        self.config.configure()
        self.pipeline = self.config.pipeline

    def _setup_args(self):
        parser = optparse.OptionParser()
        parser.add_option('-v', '--verbose', action='store_true', help='Verbose logging (debug)')
        parser.add_option('-l', '--log', help='Log to the given log file', default='logcabin.log')
        parser.add_option('-c', '--config', help='Configuration file to use', default='config.py')
        opts, args = parser.parse_args()

        # logging configuration
        logging_cfg = os.path.join(os.path.dirname(__file__), 'logging.cfg')
        self.logconfig = json.load(file(logging_cfg))
        logpath = os.path.abspath(opts.log)
        self.logconfig['handlers']['file']['filename'] = logpath

        if opts.verbose:
            # output to console at DEBUG too
            root = self.logconfig['loggers']['']
            root['handlers'].append('console')
            root['level'] = 'DEBUG'
        else:
            print 'Logging to %s' % logpath

        self.opts = opts

    def _setup_signals(self):
        self.shutdown = gevent.event.Event()
        gevent.signal(signal.SIGTERM, self._signal, 'SIGTERM')
        gevent.signal(signal.SIGINT, self._signal, 'SIGINT')

    def _signal(self, name):
        self.logger.info('%s received, shutting down' % name)
        self.shutdown.set()

    def start(self):
        """Start the pipeline."""
        self.pipeline.start()
        self.logger.info('Started pipeline: %s' % self.pipeline)

    def stop(self):
        """Stop the pipeline."""
        self.logger.info('Stopping pipeline')
        self.pipeline.stop()

    def run(self):
        """Run the whole application, then wait on termination by signal."""
        try:
            self.setup()
        except ConfigException as ex:
            self.logger.error(str(ex))
            print >>sys.stderr, str(ex)
            return False

        self.start()
        self.shutdown.wait()
        self.stop()

        self.logger.info('Shutdown complete')
        return True


if __name__ == '__main__':
    logcabin.main()



from ..event import Event
from .input import Input
import gevent
import zmq.green as zmq

class Zeromq(Input):
    """Receives from a zeromq socket.

    Creates events with the field 'data' set to the packet received.

    :param string address: zeromq address to bind on (default: `tcp://*:2120`)
    :param string mode: connect or bind (default: bind)
    :param string socket: PULL or SUB (default: PULL)
    """
    def __init__(self, address='tcp://*:2120', mode='bind', socket='PULL'):
        if mode not in ('connect', 'bind'):
            raise ValueError('mode should be connect or bind')
        if socket not in ('PULL', 'SUB'):
            raise ValueError('socket should be PULL or SUB')
        super(Zeromq, self).__init__()
        self.ctx = zmq.Context()
        self.sock = self.ctx.socket(getattr(zmq, socket))
        self.address = address
        if address == 'tcp://*':
            self.sock.bind_to_random_port(address)
        elif mode == 'connect':
            self.sock.connect(self.address)
        else:
            self.sock.bind(self.address)

    def _run(self):
        try:
            while True:
                data = self.sock.recv()
                self.logger.debug('Received: %r' % data)
                self.output.put(Event(data=data))
                gevent.sleep() # yield for other stages
        finally:
            # cleanup
            self.sock.close()



import sys
import logging

from pipeline import Pipeline

class ConfigException(Exception):
    pass

class PyConfigLoader(object):
    def __init__(self, path):
        self.logger = logging.getLogger('config')
        self.path = path

    def configure(self):
        self.logger.info('Loading configuration: %r' % self.path)
        # push implicit top-level context
        self.pipeline = Pipeline()
        with self.pipeline:
            try:
                execfile(self.path)
            except Exception as ex:
                trace = sys.exc_info()[2]
                raise ConfigException(ex), None, trace

        # pipeline has no final output queue
        self.pipeline.setup(None)
